# -*- coding: utf-8 -*-

import sys
import os
import pytest
import numpy as np
import xarray as xr
from matrixconverters.xarray2netcdf import xr2netcdf


@pytest.fixture(scope='class')
def folder():
    return os.path.dirname(__file__)


@pytest.fixture(scope='class')
def fp_dataset_with_umlaut(folder):
    fn = os.path.join(folder, 'ÄÖÜß', 'ÄÖÜß.nc4')
    return fn


@pytest.fixture(scope='class')
def fp_dataset_without_umlaut(folder):
    fn = os.path.join(folder, 'dataset_netcdf.nc4')
    return fn

@pytest.fixture(scope='class')
def ds():
    da = xr.DataArray(np.arange(9).reshape(3, 3))
    zones = xr.DataArray([100, 200, 300])
    names = xr.DataArray(['ÄÄÄ-Town', 'B-Village', 'C-City'])
    ds = xr.Dataset({'matrix': da,
                     'zone_no': zones,
                     'zone_name': names,})
    return ds


class TestSaveReadxr2netcdf:
    def test_01_save_ds(self, ds, fp_dataset_without_umlaut):
        # if the filepath containes non-ascii chars, you have to use the
        # legacy encoding in windows >= python3.6
        if sys.platform == 'win32' and sys.version_info >= (3, 6):
            sys._enablelegacywindowsfsencoding()

        xr2netcdf(ds, fp_dataset_without_umlaut)
        ds_saved = xr.open_dataset(fp_dataset_without_umlaut)
        np.testing.assert_array_equal(ds_saved.matrix, ds.matrix)
        np.testing.assert_array_equal(ds_saved.zone_no, ds.zone_no)
        np.testing.assert_array_equal(ds_saved.zone_name, ds.zone_name)

    @pytest.mark.skip(msg='netcdf does not support filepath with umlaut')
    def test_02_save_ds_with_umlaut(self, ds, fp_dataset_with_umlaut):
        # if the filepath containes non-ascii chars, you have to use the
        # legacy encoding in windows >= python3.6
        if sys.platform == 'win32' and sys.version_info >= (3, 6):
            sys._enablelegacywindowsfsencoding()

        xr2netcdf(ds, fp_dataset_with_umlaut)
        ds_saved = xr.open_dataset(fp_dataset_with_umlaut)
        np.testing.assert_array_equal(ds_saved.matrix, ds.matrix)
        np.testing.assert_array_equal(ds_saved.zone_no, ds.zone_no)
        np.testing.assert_array_equal(ds_saved.zone_name, ds.zone_name)



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 10 21:00:21 2016

@author: MaxBohnet
"""

import pytest
import os
import numpy as np
from matrixconverters.read_csv import ReadOFormat
from matrixconverters.tests.test_read_visum import folder


@pytest.fixture(scope='class')
def fn_zones(folder):
    fn = os.path.join(folder, 'zones.csv')
    return fn


@pytest.fixture(scope='class')
def fn_matrix(folder):
    fn = os.path.join(folder, 'matrix.csv')
    return fn


class TestReadCSV:
    """Test reading CSV matrices"""
    def test_01_read_csv(self, fn_zones, fn_matrix):
        """Test reading CSV-Format"""
        ds = ReadOFormat(zonefile=fn_zones,
                         matrixfile=fn_matrix)
        print(ds)
        print(ds.matrix)
        np.testing.assert_allclose(ds.matrix.sum(), 198.1)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import xarray as xr
from typing import Iterable


class ReadOFormat(xr.Dataset):
    """reads Matrix from a file
    """
    __slots__ = ()
    _target_cols_zones = ['zone_no', 'zone_name']
    _target_cols_matrix = ['origins', 'destinations', 'values']

    def __init__(self,
                 zonefile: str,
                 matrixfile: str,
                 cols_zones: Iterable[str] = None,
                 cols_matrix: Iterable[str] = None
                 ):
        """
        Parameters
        ----------
        zonefile : str
            the filepath to the file with zone names
        matrixfile : str
            the filepath to the file with the matrix values
        cols_zones : list-like, optional
        cols_matrix : list-like, optional
        """
        super().__init__()
        self.read_zones_csv(zonefile, cols_zones)
        self.read_matrix_csv(matrixfile, cols_matrix)

    def read_matrix_csv(self, filename: str, cols_matrix: Iterable[str]):
        """
        Reads a matrix from a csv-file and stores it in self['matrix']

        Parameters
        ----------
        filename : str
            the filepath of the input file
        cols_matrix : list-like
            the columns to use
        """
        target_cols = self._target_cols_matrix
        data_cols = cols_matrix
        pkey = target_cols[:2]

        da = self.read_file_to_da(data_cols, filename, target_cols, pkey)
        self['matrix'] = da['values']
        m = self['matrix']
        m.data[np.isnan(m.data)] = 0

    def read_zones_csv(self, filename: str, cols_zones: Iterable[str]):
        """
        Reads a matrix from a csv-file and stores it as coordinates


        Parameters
        ----------
        filename : str
            the filepath of the input file
        cols_matrix : list-like
            the columns to use [column_with_zone_no, column_with_zone_name]
        """
        target_cols = self._target_cols_zones
        data_cols = cols_zones
        pkey = target_cols[0]
        col_name = target_cols[1]

        da = self.read_file_to_da(data_cols, filename, target_cols, pkey)
        dims = self._target_cols_zones[:1] + self._target_cols_matrix[:2]
        # set as origins and destinations
        for dim in dims:
            renamed_da = da.rename({pkey: dim})
            self[dim] = renamed_da[dim]
            name_dim = 'name_{}'.format(dim)
            self.coords[name_dim] = xr.IndexVariable(dims=[dim],
                                                     data=renamed_da[col_name])

    def read_file_to_da(self,
                        data_cols: Iterable[str],
                        filename: str,
                        target_cols: Iterable[str],
                        pkey: str) -> xr.DataArray:
        """
        reads a file into a DataArray

        Parameters
        ----------
        data_cols : list-like
        filename : str
        target_cols : list_like
        pkey : str

        Returns
        -------
        da : xarray.DataArray
        """
        df = pd.read_csv(filename, usecols=data_cols)
        if not data_cols:
            # take the first columns of the table
            data_cols = df.columns.values[:len(target_cols)]

        # rename the columns to the target names
        msg = 'wrong number of columns specified'
        assert len(data_cols) == len(target_cols), msg
        rename_cols = dict(zip(data_cols,
                               target_cols))
        df.rename(columns=rename_cols, inplace=True)

        da = df.set_index(pkey).sort_index().to_xarray()
        return da



# -*- coding: utf-8 -*-
import xarray as xr


def xr2netcdf(dataset: xr.Dataset,
              filepath: str,
              mode: str = 'w',
              engine: str = 'netcdf4',
              compressed: bool = True,
              complevel: int = 2, ):
    """
    save dataset as netcdf-file to filepath using the given compression level

    Parameters
    ----------
    dataset : xarray-dataset
        the xarray-dataset to store
    filepath : str
        the path where the netcdf-file shold be stored
    mode : str, default='w'
        write (w) or append (a)
    engine : str (optional, default=netcdf4)
        the engine to use
    compressed : boolen (optional, default=True)
        if False, the data-variables are not compressed
    complevel : int (optional, default=2)
        the compression-level between 1 and 9
        1 is faster, 9 uses a more efficient compression, but is much slower
        2 is in general a good compromise
    """
    encoding = {}
    comp = {}
    if compressed:
        comp['zlib'] = 'True'
        comp['complevel'] = complevel

    for data_var in dataset.data_vars:
        encoding[data_var] = comp
    dataset.to_netcdf(filepath, mode=mode, engine=engine, encoding=encoding)



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 10 21:00:21 2016

@author: MaxBohnet
"""

import pytest
import os
import numpy as np
from matrixconverters.read_ptv import ReadPTVMatrix
from matrixconverters.save_ptv import SavePTV
import xarray as xr


@pytest.fixture(scope='class')
def folder():
    return os.path.dirname(__file__)


@pytest.fixture(scope='class')
def matrix_fn(folder):
    fn = os.path.join(folder, 'matrix_v_format.mtx')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_out(folder):
    fn = os.path.join(folder, 'matrix.out')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_bk_out(folder):
    fn = os.path.join(folder, 'matrix_bk_format.out')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_bi_out(folder):
    fn = os.path.join(folder, 'matrix_bi_format.out')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_bk(folder):
    fn = os.path.join(folder, 'matrix_bk_format.mtx')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_bi(folder):
    fn = os.path.join(folder, 'matrix_bi_format.mtx')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_bl_out(folder):
    fn = os.path.join(folder, 'matrix_bl_format.out')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_bl(folder):
    fn = os.path.join(folder, 'matrix_bl_format.mtx')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_o(folder):
    fn = os.path.join(folder, 'matrix_o_format.mtx')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_o_without_names(folder):
    fn = os.path.join(folder, 'matrix_o_format_without_names.mtx')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_o_with_zeros(folder):
    fn = os.path.join(folder, 'matrix_o_format_with_zeros.mtx')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_o_with_dashes(folder):
    fn = os.path.join(folder, 'matrix_o_format_with_dashes.mtx')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_e(folder):
    fn = os.path.join(folder, 'matrix_e_format.mtx')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_s(folder):
    fn = os.path.join(folder, 'matrix_s_format.mtx')
    return fn


@pytest.fixture(scope='class')
def matrix_fn_or(folder):
    fn = os.path.join(folder, 'matrix_or_format.mtx')
    return fn


class TestReadPTV:
    """Test reading PTV matrices"""
    def test_01_read_v_format(self, matrix_fn):
        """Test reading v-Format"""
        ds = ReadPTVMatrix(filename=matrix_fn)
        self.print_matrix(ds)

    def test_02_read_bk_format(self, matrix_fn_bk):
        """Test reading bk-format"""
        ds = ReadPTVMatrix(filename=matrix_fn_bk)
        self.print_matrix(ds)

    def test_02a_read_o_format(self,
                               matrix_fn_o,
                               matrix_fn_or,
                               matrix_fn_o_without_names,
                               matrix_fn_o_with_zeros,
                               matrix_fn_o_with_dashes):
        """Test reading bk-format"""
        zone_no = [100, 200, 300, 400]
        ds = ReadPTVMatrix(filename=matrix_fn_o)
        self.print_matrix(ds)
        np.testing.assert_equal(ds.zone_no.data, zone_no)

        ds = ReadPTVMatrix(filename=matrix_fn_or)
        self.print_matrix(ds)
        np.testing.assert_equal(ds.zone_no.data, zone_no)

        ds = ReadPTVMatrix(filename=matrix_fn_o_without_names)
        self.print_matrix(ds)
        np.testing.assert_equal(ds.zone_no.data, zone_no)

        ds = ReadPTVMatrix(filename=matrix_fn_o_with_zeros)
        self.print_matrix(ds)
        np.testing.assert_equal(ds.zone_no.data, zone_no)
        np.testing.assert_equal((ds.matrix.data != 0).sum(), 2,
                                err_msg='only 2 values should be different to 0')

        ds = ReadPTVMatrix(filename=matrix_fn_o_with_dashes)
        self.print_matrix(ds)
        np.testing.assert_equal(ds.zone_no.data, zone_no)
        np.testing.assert_equal((ds.matrix.data != 0).sum(), 1,
                                err_msg='only 1 values should be different to 0')


    def test_02b_read_e_s_format(self, matrix_fn_e, matrix_fn_s):
        """Test reading bk-format"""
        ds = ReadPTVMatrix(filename=matrix_fn_e)
        self.print_matrix(ds)

        ds = ReadPTVMatrix(filename=matrix_fn_s)
        self.print_matrix(ds)

    def test_02c_read_bi_format(self, matrix_fn_bi):
        """Test reading bki-format"""
        ds = ReadPTVMatrix(filename=matrix_fn_bi)
        self.print_matrix(ds)

    def print_matrix(self, ds):
        print(ds)
        print('Histogram of travel times')
        print(np.histogram(ds.matrix.data, bins=range(0, 60, 10)))
        sum_before = ds.matrix.data.sum()
        print(sum_before)

    def test_03_save_v_format(self, matrix_fn_bk, matrix_fn_out):
        """Test writing v-format"""
        ds = ReadPTVMatrix(filename=matrix_fn_bk)
        print(np.histogram(ds.matrix.data, bins=range(0, 60, 10)))
        sum_before = ds.matrix.data.sum()
        print(sum_before)
        s = SavePTV(ds)
        s.savePTVMatrix(file_name=matrix_fn_out,
                        file_type='VN', )
        ds2 = ReadPTVMatrix(filename=matrix_fn_out)
        print(np.histogram(ds2.matrix.data, bins=range(0, 60, 10)))
        sum_after = ds2.matrix.data.sum()
        print(sum_after)
        np.testing.assert_almost_equal(sum_before, sum_after, decimal=5)

    def test_03b_save_vm_format(self, matrix_fn_bk, matrix_fn_out):
        """Test writing vm-format"""
        ds = ReadPTVMatrix(filename=matrix_fn_bk)
        sum_before = ds.matrix.data.sum()
        s = SavePTV(ds)
        s.savePTVMatrix(file_name=matrix_fn_out,
                        file_type='VM', )
        ds2 = ReadPTVMatrix(filename=matrix_fn_out)
        sum_after = ds2.matrix.data.sum()
        np.testing.assert_almost_equal(sum_before, sum_after, decimal=5)

    def test_05_save_bk_format(self, matrix_fn_bk, matrix_fn_bk_out):
        """Test writing bk-format"""
        ds = ReadPTVMatrix(filename=matrix_fn_bk)
        print(np.histogram(ds.matrix.data, bins=range(8)))
        sum_before = ds.matrix.data.sum()
        print(sum_before)
        s = SavePTV(ds)
        s.savePTVMatrix(file_name=matrix_fn_bk_out,
                        file_type='BK', )
        ds2 = ReadPTVMatrix(filename=matrix_fn_bk_out)
        print(np.histogram(ds2.matrix.data, bins=range(8)))
        sum_after = ds2.matrix.data.sum()
        print(sum_after)
        assert sum_before == sum_after
        #  test if it raises Exception, if zone numbers do not match the matrix
        del ds['zone_name']
        ds['zone_no'] = xr.IndexVariable(('zone_no', ), ds['zone_no'][:3])
        with pytest.raises(AssertionError):
            SavePTV(ds).savePTVMatrix(file_name=matrix_fn_bk_out,
                        file_type='BK', )

    def test_05a_save_bi_format(self, matrix_fn_bi, matrix_fn_bi_out):
        """Test writing bi-format"""
        ds = ReadPTVMatrix(filename=matrix_fn_bi)
        print(np.histogram(ds.matrix.data, bins=range(8)))
        sum_before = ds.matrix.data.sum()
        print(sum_before)
        s = SavePTV(ds)
        s.savePTVMatrix(file_name=matrix_fn_bi_out, file_type='BI')
        ds2 = ReadPTVMatrix(filename=matrix_fn_bi_out)
        print(np.histogram(ds2.matrix.data, bins=range(8)))
        sum_after = ds2.matrix.data.sum()
        print(sum_after)
        assert sum_before == sum_after

    def test_05b_save_bi_format_zero(self, matrix_fn_bi, matrix_fn_bi_out):
        """Test writing bi-format with zero flag"""
        ds = ReadPTVMatrix(filename=matrix_fn_bi)
        ds.matrix.data[:] = 0
        print(np.histogram(ds.matrix.data, bins=range(8)))
        sum_before = ds.matrix.data.sum()
        print(sum_before)
        s = SavePTV(ds)
        s.savePTVMatrix(file_name=matrix_fn_bi_out, file_type='BI')
        ds2 = ReadPTVMatrix(filename=matrix_fn_bi_out)
        print(np.histogram(ds2.matrix.data, bins=range(8)))
        sum_after = ds2.matrix.data.sum()
        print(sum_after)
        assert sum_before == sum_after

    def test_06_save_o_format(self, matrix_fn_bk, matrix_fn_out):
        """Test writing v-format"""
        ds = ReadPTVMatrix(filename=matrix_fn_bk)
        print(np.histogram(ds.matrix.data, bins=range(0, 60, 10)))
        sum_before = ds.matrix.data.sum()
        print(sum_before)
        s = SavePTV(ds)
        s.savePTVMatrix(file_name=matrix_fn_out,
                        file_type='O', )
        ds2 = ReadPTVMatrix(filename=matrix_fn_out)
        print(np.histogram(ds2.matrix.data, bins=range(0, 60, 10)))
        sum_after = ds2.matrix.data.sum()
        print(sum_after)
        np.testing.assert_almost_equal(sum_before, sum_after, decimal=5)

    def test_07_save_bl_format(self, matrix_fn_bl, matrix_fn_bl_out):
        """Test writing bk-format"""
        ds = ReadPTVMatrix(filename=matrix_fn_bl)
        print(np.histogram(ds.matrix.data, bins=range(8)))
        sum_before = ds.matrix.data.sum()
        print(sum_before)
        ds['zone_name'][:] = ['A', 'B', 'C', 'D', 'E']
        ds['zone_names2'][:] = ['ÄÄ', 'ÜÜ', 'Öß€']
        ds = ds.assign_coords(zone_no2=ds['zone_no2'] * 100)
        s = SavePTV(ds)
        s.savePTVMatrix(file_name=matrix_fn_bl_out,
                        file_type='BK', )
        ds2 = ReadPTVMatrix(filename=matrix_fn_bl_out)
        print(np.histogram(ds2.matrix.data, bins=range(8)))
        sum_after = ds2.matrix.data.sum()
        print(sum_after)
        assert sum_before == sum_after
        np.testing.assert_array_equal(ds.zone_no, ds2.zone_no)
        np.testing.assert_array_equal(ds.zone_no2, ds2.zone_no2)
        np.testing.assert_array_equal(ds.zone_name, ds2.zone_name)
        np.testing.assert_array_equal(ds.zone_names2, ds2.zone_names2)

    def test_08_save_simple_ds(self, matrix_fn_bk_out):
        da = xr.DataArray(np.arange(9).reshape(3, 3))
        zones = xr.DataArray([100, 200, 300])
        names = xr.DataArray(['A-Town', 'B-Village', 'C-City'])
        ds = xr.Dataset({'matrix': da,
                         'zone_no': zones,
                         'zone_name': names,})
        s = SavePTV(ds)
        s.savePTVMatrix(matrix_fn_bk_out)

        ds_saved = ReadPTVMatrix(matrix_fn_bk_out)
        print(ds_saved)
        np.testing.assert_array_equal(ds_saved.matrix, da)
        np.testing.assert_array_equal(ds_saved.zone_no, zones)
        np.testing.assert_array_equal(ds_saved.origins, zones)
        np.testing.assert_array_equal(ds_saved.destinations, zones)
        np.testing.assert_array_equal(ds_saved.zone_name, names)
        np.testing.assert_array_equal(ds_saved.zone_names2, names)



# -*- coding: utf-8 -*-

import zlib
import numpy as np
import xarray as xr
from typing import BinaryIO


def write_line(f, text, length):
    f.write(text.ljust(length) + b"\r\n")


class SavePTV(object):

    def __init__(self, ds: xr.Dataset):
        """
        Parameters
        ----------
        ds: xarray-Dataset
            the dataset that should be saved
        """
        self.ds = ds

    @staticmethod
    def write_u1(f: BinaryIO, val: int):
        """write 1-byte number to file"""
        f.write(np.array(val, dtype="u1").tostring())

    @staticmethod
    def write_i2(f: BinaryIO, val: int):
        """write short number to file"""
        f.write(np.array(val, dtype="i2").tostring())

    @staticmethod
    def write_i4(f: BinaryIO, val: int):
        """write int number to file"""
        f.write(np.array(val, dtype="i4").tostring())

    @staticmethod
    def write_f4(f: BinaryIO, val: float):
        """write float to file"""
        f.write(np.array(val, dtype="f4").tostring())

    @staticmethod
    def write_f8(f: BinaryIO, val: float):
        """write double to file"""
        f.write(np.array(val, dtype="f8").tostring())

    def write_utf16(self, f: BinaryIO, val: xr.DataArray):
        """write string as utf16 encoded string to file"""
        unencoded_str = val.data.tolist()
        self.write_i4(f, len(unencoded_str))
        utf16_str = unencoded_str.encode('UTF-16LE')
        f.write(utf16_str)

    def savePTVMatrix(self,
                      file_name: str,
                      file_type: str = 'BK',
                      version: float = 0.11,
                      ):
        """
        save array im PTV-Format

        Parameters
        ----------
        file_name : string
            output file_name
        file_type : str
            Type of PTV-File Format
        version : str
            version number
        """

        if file_type.startswith("B"):
            self._write_format_b(file_name, file_type=file_type)
        elif file_type.startswith("V"):
            self._write_format_v(file_name, file_type)
        elif file_type.startswith("O"):
            self._write_format_o(file_name, file_type)

    def _write_format_v(self, fn: str, file_type: str):
        dtypes = {'float32': "Y4", 'float64': "Y5",
                  'int32': "Y3", 'int16': "Y2"}
        outformat = {'float32': "%0.3f", 'float64': "%0.6f",
                     'int32': "%d", 'int16': "%d"}
        with open(fn, "w") as f:
            ZeitVon = getattr(self.ds.attrs, 'ZeitVon', 0)
            ZeitBis = getattr(self.ds.attrs, 'ZeitBis', 24)
            Faktor = getattr(self.ds.attrs, 'Faktor', 1)
            VMAktKennung = getattr(self.ds.attrs, 'VMAktKennung', 0)
            dtype = self.ds.matrix.dtype
            f.write("$%s, %s\n" % (file_type, dtypes[str(dtype)]))
            if "M" in file_type:
                f.write("* Verkehrsmittelkennung:\n %d \n" % VMAktKennung)
            if "N" not in file_type:
                f.write("* Zeitintervall:\n {v} {b} \n".format(
                    v=ZeitVon,
                    b=ZeitBis))
                f.write("* Faktor:\n %d \n" % Faktor)
            zone_no = self.ds.zone_no.data
            n_zones = len(zone_no)
            f.write("* Anzahl Bezirke:\n %d\n" % n_zones)
            f.write("* Bezirksnummern\n")
            for i in range(0, n_zones, 10):
                f.write("\t".join(map(lambda x:
                                      str(x),
                                      zone_no[i: i + 10])) + "\n")
            f.write("* Matrixwerte\n")
            values = self.ds.matrix.data
            fmt = outformat[str(dtype)]
            for i in range(n_zones):
                no = zone_no[i]
                f.write("* %s \n" % no)
                for j in range(0, n_zones, 10):
                    values[i, j: j + 10].tofile(f, sep=' ', format=fmt)
                    f.write('\n')

            if 'zone_name' in self.ds:
                zone_name = self.ds.zone_name.data
                fmt = '{no} "{name}"\n'
                for i in range(n_zones):
                    f.write(fmt.format(no=zone_no[i], name=zone_name[i]))

    def _write_format_o(self, fn: str, file_type: str):
        dtypes = {'float32': "Y4", 'float64': "Y5",
                  'int32': "Y3", 'int16': "Y2"}

        with open(fn, "w") as f:
            ZeitVon = self.ds.attrs.get('ZeitVon', 0)
            ZeitBis = self.ds.attrs.get('ZeitBis', 24)
            Faktor = self.ds.attrs.get('Faktor', 1)
            VMAktKennung = self.ds.attrs.get('VMAktKennung', 0)
            dtype = self.ds.matrix.dtype
            f.write("$%s, %s\n" % (file_type, dtypes[str(dtype)]))
            if "M" in file_type:
                f.write("* Verkehrsmittelkennung:\n %d \n" % VMAktKennung)
            if "N" not in file_type:
                f.write("* Zeitintervall:\n {v} {b} \n".format(
                    v=ZeitVon,
                    b=ZeitBis))
                f.write("* Faktor:\n %d \n" % Faktor)
            df = self.ds.matrix.to_dataframe().iloc[:, 0]
            df_larger_0 = df[df > 0]
            f.write("* Matrixwerte\n")
            df_larger_0.to_csv(f, sep=' ', header=False)

            zone_no = self.ds.zone_no.data
            n_zones = self.ds.dims['zone_no']

            if 'zone_name' in self.ds:
                f.write('* Netzobjektnamen\n')
                f.write('$NAMES\n')
                zone_name = self.ds.zone_name.data
                fmt = '{no} "{name}"\n'
                for i in range(n_zones):
                    f.write(fmt.format(no=zone_no[i], name=zone_name[i]))

    def _write_format_b(self, fn: str, file_type: str = 'K'):
        m = self.ds.matrix.data
        with open(fn, "wb") as f:

            n_zones = m.shape[0]
            n_cols = m.shape[1]

            idlength = 3
            self.write_i2(f, idlength)
            is_square = n_zones == n_cols
            if is_square:
                if file_type == 'BI':
                    idvalue = b'$BI'
                else:
                    idvalue = b'$BK'
            else:
                idvalue = b'$BL'
            compression_type = chr(idvalue[2])
            f.write(idvalue)

            vartype = 5

            datatypes = {'f': 4, 'd': 5, 'l': 3, 'h': 2, }
            typecodes = {'f': 'f4', 'd': 'f8', 'l': 'i4', 'h': 'i2', }
            datatype = datatypes.get(m.dtype.char, 5)
            typecode = typecodes.get(m.dtype.char, 'f8')

            header_lines = ['']
            ZeitVon = self.ds.attrs.get('ZeitVon', 0)
            ZeitBis = self.ds.attrs.get('ZeitBis', 24)
            Faktor = self.ds.attrs.get('Faktor', 1)
            VMAktKennung = self.ds.attrs.get('VMAktKennung', 0)
            roundproc = self.ds.attrs.get('roundproc', 0)

            # write header
            header_lines.append("Muuli matrix in packed binary format.")
            header_lines.append("Zones: {} ".format(n_zones))
            header_lines.append("VarType: {} ".format(vartype))
            header_lines.append("Total sum: {:.6f} ".format(m.sum()))

            diagsum = np.diag(m).sum()
            header_lines.append("Diagonal sum: {:6f} ".format(diagsum))
            header_lines.append("Transport mode: {} ".format(VMAktKennung))
            header_lines.append("from: {:.2f} ".format(ZeitVon))
            header_lines.append("to: {:.2f} ".format(ZeitBis))
            header_lines.append("Factor: {:6f} ".format(Faktor))
            header_lines.append("")

            header = '\r\n'.join(header_lines).encode('utf8')
            header_length = len(header)
            self.write_i2(f, header_length)
            f.write(header)

            # write additional infos
            self.write_i4(f, VMAktKennung)
            self.write_f4(f, ZeitVon)
            self.write_f4(f, ZeitBis)
            self.write_f4(f, Faktor)
            self.write_i4(f, n_zones)
            self.write_i2(f, datatype)
            self.write_u1(f, roundproc)

            # columns
            if compression_type != 'I':
                self.write_i4(f, n_cols)
            # zone numbers for rows ...
            zone_no = self.ds.zone_no
            assert len(zone_no) == n_zones,\
                   'zone_no with length {d} does not match matrix with {n} rows'\
                .format(d=len(zone_no), n=n_zones)
            f.write(np.array(zone_no.data).astype("i4").tostring())
            # ... and columns (not for $BI)
            if compression_type != 'I':
                zone_cols = getattr(self.ds, 'zone_no2', zone_no)
                assert len(zone_cols) == n_cols,\
                    'zone_cols with length {d} does not match matrix with {n} columns'\
                    .format(d=len(zone_cols), n=n_cols)
                f.write(np.array(zone_cols.data).astype("i4").tostring())

                # zone_names (not for $BI):
                zone_names = getattr(self.ds, 'zone_name', None)
                if zone_names is None:
                    zone_names = [xr.DataArray('') for i in range(n_zones)]
                assert len(zone_names) == n_zones,\
                    'zone_names with length {d} does not match matrix with {n} rows'\
                    .format(d=len(zone_names), n=n_zones)
                # for rows
                for zone_name in zone_names:
                    self.write_utf16(f, zone_name)
                # for columns
                zone_names2 = getattr(self.ds, 'zone_names2', None)
                if zone_names2 is None:
                    if is_square:
                        zone_names2 = zone_names
                    else:
                        zone_names2 = [xr.DataArray('') for i in range(n_cols)]
                assert len(zone_names2) == n_cols,\
                    'zone_names2 with length {d} does not match matrix with {n} columns'\
                    .format(d=len(zone_names2), n=n_cols)
                for zone_name in zone_names2:
                    self.write_utf16(f, zone_name)

            allnull = int(not np.any(m))
            self.write_u1(f, allnull)
            self.write_f8(f, diagsum)
            if not allnull:
                data = m.astype(typecode)
                rowsums = data.sum(1).astype('f8')
                colsums = data.sum(0).astype('f8')
                for i in range(n_zones):
                    row = data[i]
                    compressed = zlib.compress(row.tostring())
                    self.write_i4(f, len(compressed))
                    f.write(compressed)
                    if compression_type != 'L':
                        self.write_f8(f, rowsums[i])
                        self.write_f8(f, colsums[i])
                if not is_square:
                    # for $BL-format, the row and colsums are written as vector
                    f.write(rowsums.tostring())
                    f.write(colsums.tostring())

    def savePSVMatrix(self,
                      file_name: str,
                      ftype: str = "CC",
                      max_width: int = 1000):
        """
        exports array in PSV-Format
        """
        m = self.ds.matrix.data
        if not m.ndim == 2:
            raise ValueError('matrix has to have 2 dimensions')
        rows, cols = m.shape
        zones = self.ds.get('zone_no1', range(1, rows + 1))

        with open(file_name, "w") as f:
            Za = max(rows, cols)
            Zi = min(rows, cols)
            f.write("%s; Za %s; Zi %s;\n" % (ftype, Za, Zi))
            if ftype == "CC":
                for i in range(rows):
                    # teste, ob Zeile nicht nur aus Nullen besteht
                    if m[i].any():
                        # addiere 1, um Zellennummer 0 zu vermeiden
                        row = str(zones[i])
                        for j in range(cols):
                            if len("%s %s %s" % (row, zones[j],
                                                 m[i, j])) >= max_width:
                                row += "\n"
                                f.write(row)
                                row = str(zones[i])
                            if m[i, j] != 0:
                                row += " %s %s " % (zones[j], m[i, j])
                        row += "\n"
                        f.write(row)

            elif ftype == "CN":
                for i in range(rows):
                    # teste, ob Zeile nicht nur aus Nullen besteht
                    if m[i].any():
                        row = str(zones[i])
                        for j in range(cols):
                            if len("%s %s" % (row, m[i, j])) >= max_width:
                                row += "\n"
                                f.write(row)
                                row = str(zones[i])
                            row += " " + str(m[i, j])
                        row += "\n"
                        f.write(row)
            else:
                raise TypeError("Kann Typ %s nicht schreiben" % ftype)



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 10 21:00:21 2016

@author: MaxBohnet
"""

import pytest
import os
import numpy as np
from matrixconverters.read_ptv import ReadPTVMatrix
from matrixconverters.save_ptv import SavePTV
from matrixconverters.tests.test_read_visum import folder, matrix_fn


@pytest.fixture(scope='class')
def matrix_cc(folder):
    fn = os.path.join(folder, 'matrix.cc')
    return fn


@pytest.fixture(scope='class')
def matrix_cc_out(folder):
    fn = os.path.join(folder, 'matrix_out.cc')
    return fn


@pytest.fixture(scope='class')
def matrix_cn(folder):
    fn = os.path.join(folder, 'matrix.cn')
    return fn

@pytest.fixture(scope='class')
def matrix_cn_out(folder):
    fn = os.path.join(folder, 'matrix_out.cn')
    return fn


class TestSavePSV:
    """Test saving PSV matrices"""
    def test_01_write_cc(self, matrix_fn, matrix_cc_out):
        """Test writing CC-Format"""
        ds = ReadPTVMatrix(filename=matrix_fn)
        s = SavePTV(ds)
        s.savePSVMatrix(file_name=matrix_cc_out,
                        ftype='CC', )

    def test_02_write_cn(self, matrix_fn, matrix_cn_out):
        """Test writing CC-Format"""
        ds = ReadPTVMatrix(filename=matrix_fn)
        s = SavePTV(ds)
        s.savePSVMatrix(file_name=matrix_cn_out,
                            ftype='CN', )



# -*- coding: utf-8 -*-

import zlib
from collections import deque
import numpy as np
import xarray as xr
from typing import TextIO, BinaryIO, IO, Iterable, Tuple, List


class ReadPTVMatrix(xr.Dataset):
    """reads PTV-Matrix from file
    the matrix-type is determined by the header
    reades $V, $O, $E, and the binary types $BI, $BK and $B
    """
    __slots__ = ()

    def __init__(self, filename: str):
        """
        read a PTV-Matrix and return it as an xarray-dataset

        Parameters
        ----------
        filename: str
            path to file on disk
        """
        super().__init__()
        self.attrs['fn'] = filename
        self.attrs['ZeitVon'] = 0
        self.attrs['ZeitBis'] = 0
        self.attrs['Faktor'] = 1
        self.attrs['VMAktKennung'] = 0
        self.attrs['AnzBezeichnerlisten'] = 1
        self.attrs['roundproc'] = 1

        with self._openfile(mode='rb') as f:
            line = f.readline().strip()

            errmsg = 'Matrix type not recognised form Header {}'.format(line)

        if line.startswith(b'$') and len(line) > 1:
            # text format
            matrix_type = chr(line[1])
            if line.startswith(b"$"):
                read_methods = {
                    'V': self.readPTVMatrixV,
                    'O': self.readPTVMatrixO,
                    'E': self.readPTVMatrixE,
                    'S': self.readPTVMatrixE,
                }
                read = read_methods.get(matrix_type)
                if not read:
                    raise ValueError(errmsg)
                read()
            else:
                raise ValueError(errmsg)
        else:
            # assume it is the binary format
            self.readPTVMatrixB()

    def readPTVMatrixO(self):
        """read a file in O-Format"""
        with self._openfile(mode="r") as f:
            rows = self.read_header(f)
        loc = self.matrix.loc
        for row in rows:
            r = row.split()
            fr = int(r[0])
            to = int(r[1])
            value = self.float_nan(r[2])
            loc[fr, to] = value

    def readPTVMatrixE(self):
        """read a file in E-Format"""
        with self._openfile(mode="r") as f:
            rows = self.read_header(f)
        loc = self.matrix.loc
        for row in rows:
            cols = row.lstrip('-').split()
            fr = int(cols[0])
            for i in range(1, len(cols), 2):
                to = int(cols[i])
                value = float(cols[i + 1])
                loc[fr, to] = value

    def read_header(self, f: TextIO) -> deque:
        """
        read the header

        Parameters
        ----------
        f : file-obj

        Returns
        -------
        rows: deque
        """
        line = f.readline()
        MatrixTyp = line.split("$")[-1].split(";")[0]
        if "M" in MatrixTyp:
            self.attrs['VMAktKennung'] = self.read_value(f)

        if "N" not in MatrixTyp:
            ZeitVon, ZeitBis, Faktor = self.read_values(f, 3)
            self.attrs['ZeitVon'] = ZeitVon
            self.attrs['ZeitBis'] = ZeitBis
            self.attrs['Faktor'] = Faktor

        rows, line = self.read_values_in_o_format_to_list(f)
        self.read_names_o_format(f, line, rows)
        return rows

    def _openfile(self, mode: str = 'r', encoding: str = 'latin1') -> IO:
        """
        open the file with the according open method

        Parameters
        ----------
        mode : str, optional(default='r')
            read, write or append
        encoding : str, optional(default='latin1')

        Returns
        -------
        open file-handler
        """
        if mode.endswith('b'):
            encoding = None
        return open(self.attrs['fn'], mode=mode, encoding=encoding)

    def readPTVMatrixV(self):
        """read a file in V-Format"""
        with self._openfile() as f:
            line = f.readline()
            MatrixTyp = line.split("$")[-1].split(";")[0]
            if not MatrixTyp.startswith("V"):
                print("Keine Matrix im V-Format!")
                raise TypeError
            if "M" in MatrixTyp:
                self.attrs['VMAktKennung'] = self.read_value(f)

            if "N" not in MatrixTyp:
                ZeitVon, ZeitBis, Faktor = self.read_values(f, 3)
                self.attrs['ZeitVon'] = ZeitVon
                self.attrs['ZeitBis'] = ZeitBis
                self.attrs['Faktor'] = Faktor

            n_zones = self.read_value(f)
            self.create_zones(n_zones)
            self.read_values_to_array(f, self.zone_no)

            self.create_matrix(n_zones)
            self.read_values_to_array(f, self.matrix)
            self.create_zone_names(n_zones)
            self.read_names(f, self.zone_name)

    def create_zone_names(self,
                          n_zones: int,
                          name: str = 'zone_name',
                          dim: str = 'zone_no'):
        coord = getattr(self, dim).data
        self[name] = xr.DataArray(
            np.empty((n_zones, ), dtype='O'),
            coords=(coord, ),
            dims=(dim,),
            name=name,)

    def create_matrix(self,
                      n_zones: int,
                      n_cols: int = None,
                      dtype: str = 'f8'):
        n_cols = n_cols or n_zones
        origins = self.zone_no
        destinations = getattr(self, 'zone_no2', origins)
        self['matrix'] = xr.DataArray(
            np.zeros((n_zones, n_cols), dtype=dtype),
            coords=(origins, destinations),
            dims=('origins', 'destinations'),
            name='matrix',)

    def create_zones(self,
                     n_zones: int,
                     name: str = 'zone_no',
                     dim: str = 'zones'):
        self.coords[name] = xr.DataArray(np.arange(n_zones, dtype='i4'),
                                         dims=(dim,),)

    def read_names(self, f: TextIO, arr: xr.DataArray):
        """Read the zone names"""
        line = f.readline()
        while line.startswith("*") or line == "\n":
            line = f.readline()
        if line:
            if line.upper().startswith('$NAMES'):

                line = f.readline().strip()
                while line:
                    row = line.split(' "')
                    zone_no = int(row[0])
                    name = row[1].strip('"')
                    arr.loc[zone_no] = name
                    line = f.readline().strip()

    def read_names_o_format(self,
                            f: TextIO,
                            line: str,
                            rows: Iterable[str]):
        """Read the zone names"""
        names = deque()
        while line.startswith("*") or line == "\n":
            line = f.readline()
        if line:
            if line.upper().startswith('$NAMES'):

                line = f.readline().strip()
                while line:
                    row = line.split(' "')
                    zone_no = int(row[0])
                    name = row[1].strip('"')
                    names.append((zone_no, name))
                    line = f.readline().strip()
        if names:
            n_zones = len(names)
            self.create_zones(n_zones)
            for i, (zone_no, name) in enumerate(names):
                self.zone_no.data[i] = zone_no
            self.create_zone_names(n_zones)
            for i, (zone_no, name) in enumerate(names):
                self.zone_name.loc[zone_no] = name
        else:
            # get zone_no from unique zones in the data
            zones = []
            for row in rows:
                r = row.split()
                zones.extend(r[:2])
            zone_no = np.unique(np.array(zones, dtype=int))
            n_zones = len(zone_no)
            self.create_zones(n_zones)
            self.zone_no[:] = zone_no
        self.create_matrix(n_zones)

    def readPTVMatrixB(self):
        """
        Read a binary PTV Matrix
        """
        with self._openfile(mode="rb") as f:
            f.seek(0, 0)
            idlength = self.read_i2(f)
            idvalue = f.read(idlength)
            compression_type = chr(idvalue[2])
            header = self.read_i2(f)
            headervalue = f.read(header)
            transportvalue = self.read_i4(f)
            starttime = self.read_f4(f)
            endtime = self.read_f4(f)
            factor = self.read_f4(f)
            n_zones = self.read_i4(f)
            data_type = self.read_i2(f)
            roundproc = self.read_u1(f)
            if roundproc > 1:
                raise IOError("Flag of round procedure doesn't exist.")

            self.attrs['ZeitVon'] = starttime
            self.attrs['ZeitBis'] = endtime
            self.attrs['Faktor'] = factor
            self.attrs['roundproc'] = roundproc
            self.attrs['VMAktKennung'] = transportvalue
            self.attrs['AnzBezeichnerlisten'] = 1

            self.create_zones(n_zones)

            data_types = {2: 'i2', 3: 'i4', 4: 'f4', 5: 'f8'}
            dtype = data_types.get(data_type, 'f8')

            if compression_type == 'I':
                n_cols = n_zones
                self.create_matrix(n_zones, dtype=dtype)

                self.zone_no.data[:] = np.frombuffer(
                    f.read(n_zones * 4), dtype="i4")
            else:
                # BK or BL
                n_cols = self.read_i4(f)
                if compression_type == 'L':
                    dim = 'destinations'
                else:
                    dim = 'zones'
                self.create_zones(n_cols, name='zone_no2', dim=dim)

                self.zone_no.data[:] = np.frombuffer(
                    f.read(n_zones * 4), dtype="i4")
                self.zone_no2.data[:] = np.frombuffer(
                    f.read(n_cols * 4), dtype="i4")
                self.create_matrix(n_zones, n_cols=n_cols, dtype=dtype)

                # read zone names for rows
                self.create_zone_names(n_zones)
                for i in range(n_zones):
                    self.zone_name.data[i] = self.read_utf16(f)

                # read zone names for columns
                self.create_zone_names(n_cols, name='zone_names2',
                                       dim='zone_no2')
                for i in range(n_cols):
                    self.zone_names2.data[i] = self.read_utf16(f)

            # all null values=
            allnull = self.read_u1(f)
            if allnull > 1:
                raise IOError("Flag of allnull not correctly set.")

            # read the results
            rowsums = np.empty((n_zones), dtype='f8')
            colsums = np.empty((n_cols), dtype='f8')
            if not allnull:
                self.attrs['diagsum'] = self.read_f8(f)
                for i in range(n_zones):
                    len_chunk = self.read_i4(f)
                    unpacked = zlib.decompress(f.read(len_chunk))
                    self.matrix[i] = np.frombuffer(unpacked, dtype=dtype)
                    if compression_type < 'L':
                        # for this format row and colsums are
                        # written at each line
                        rowsums[i] = self.read_f8(f)
                        colsums[i] = self.read_f8(f)

                if compression_type >= 'L':
                    # for this format the row and colsums are written
                    # as a vector each
                    rowsums[:] = np.frombuffer(f.read(8*n_zones), dtype="f8")
                    colsums[:] = np.frombuffer(f.read(8*n_cols), dtype="f8")
            # all null
            else:
                self.attrs['diagsum'] = 0
                self.matrix[:] = 0
                rowsums[:] = 0
                colsums[:] = 0

        # check row and colsums and diagonal
        np.testing.assert_allclose(self.matrix.sum(axis=1), rowsums)
        np.testing.assert_allclose(self.matrix.sum(axis=0), colsums)
        np.testing.assert_allclose(np.diag(self.matrix).sum(),
                                   self.attrs['diagsum'])

    def read_utf16(self, f: BinaryIO) -> str:
        """read utf16( encoded string from file at current position"""
        n_chars = self.read_i4(f)
        return f.read(n_chars * 2).decode('utf16')

    @staticmethod
    def read_f4(f: BinaryIO) -> float:
        """read float from file at current position"""
        return np.frombuffer(f.read(4), dtype="f4")[0]

    @staticmethod
    def read_f8(f: BinaryIO) -> float:
        """read double from file at current position"""
        return np.frombuffer(f.read(8), dtype="f8")[0]

    @staticmethod
    def read_u1(f: BinaryIO) -> int:
        """read byte from file at current position"""
        return np.frombuffer(f.read(1), dtype="u1")[0]

    @staticmethod
    def read_i2(f: BinaryIO) -> int:
        """read short integer from file at current position"""
        return np.frombuffer(f.read(2), dtype="i2")[0]

    @staticmethod
    def read_i4(f: BinaryIO) -> int:
        """read integer from file at current position"""
        return np.frombuffer(f.read(4), dtype="i4")[0]

    def read_value(self, f: TextIO) -> int:
        """read a single value from the file as integer"""
        line = f.readline().strip()
        while not line or line.startswith("*"):
            line = f.readline()
        value = int(line)
        return value

    def read_values_to_array(self,
                             f: TextIO,
                             arr: xr.DataArray,
                             sep: str = ' '):
        """read values from the file into a DataArray"""
        flat_arr = arr.data.ravel()
        n_total = len(flat_arr)
        pos_from = 0
        while pos_from < n_total:
            line = f.readline()
            while line.startswith("*"):
                line = f.readline()
            row = np.fromstring(line, sep=sep, dtype=arr.dtype)
            pos_to = pos_from + len(row)
            flat_arr[pos_from:pos_to] = row
            pos_from = pos_to

    def read_values_in_o_format_to_list(self, f: TextIO) -> Tuple[deque, str]:
        """read values to a list"""
        rows = deque()
        for line in f:
            row = line.strip()
            if not row or row.startswith("*"):
                continue
            if row.startswith('$'):
                return rows, row
            rows.append(row)
        return rows, row

    def read_values(self, f: TextIO, n_values: int) -> List[float]:
        values = []
        found = 0
        while found < n_values:
            line = f.readline()
            while line.startswith("*"):
                line = f.readline()
            values += map(lambda x: self.float_nan(x), line.strip().split())
            found = len(values)
        return values

    @staticmethod
    def float_nan(x: str) -> float:
        """convert str to float, handling '-' as 0.0"""
        try:
            return float(x)
        except ValueError as err:
            if x == '-':
                return 0.0
            raise err




# -*- coding: utf-8 -*-
"""
Created on Fri Jun 10 20:33:08 2016

@author: MaxBohnet
"""

from setuptools import setup, find_packages
from cythoninstallhelpers.get_version import get_version
from cythoninstallhelpers.make_cython_extensions import make_extensions


ext_modnames = []

package_name = "matrixconverters"
version = get_version(package_name, __file__)

setup(
    name=package_name,
    version=version,
    description="package to read and write PTV-Visum Matrix Formats",
    url='https://maxbo.github.io/cythonarrays/',
    author='Max Bohnet',
    author_email='bohnet@ggr-planung.de',
    classifiers=[
        # How mature is this project? Common values are
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
         'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Cython',

    ],
    keywords='Visum PTV Matrix numpy xarray',

    packages=find_packages('src', exclude=['ez_setup']),
    package_dir={'': 'src'},
    package_data={'': ['*.pxd', '*.mtx', '*.csv']},
    include_package_data=True,
    zip_safe=False,
    data_files=[
        ],

    extras_require=dict(
        extra=[],
        docs=[
            'z3c.recipe.sphinxdoc',
            'sphinxcontrib-requirements'
        ],
        test=[]
    ),
    setup_requires=['pytest-runner', ],
    tests_require=['pytest', ],

    install_requires=[
        'numpy',
        'pandas',
        'xarray',
        'netCDF4',
        'cythoninstallhelpers',
    ],
    ext_modules=make_extensions(ext_modnames),
)